@@ -18,46 +18,41 @@ var log = Log
 func Mutate(ks *servingv1alpha1.KnativeServing, c client.Client) error {
 	stages := []func(*servingv1alpha1.KnativeServing, client.Client) error{
 		ingressClass,
-		egress,
 		ingress,
 		configureLogURLTemplate,
 		ensureCustomCerts,
 		imagesFromEnviron,
+		defaultToHa,
 	}
 	for _, stage := range stages {
 		if err := stage(ks, c); err != nil {
-			return err
+			return fmt.Errorf("failed to mutate KnativeServing: %w", err)
 		}
 	}
 	return nil
 }
 
+func defaultToHa(ks *servingv1alpha1.KnativeServing, c client.Client) error {
+	if ks.Spec.HighAvailability == nil {
+		ks.Spec.HighAvailability = &servingv1alpha1.HighAvailability{
+			Replicas: 2,
+		}
+	}
+
+	return nil
+}
+
 func ingressClass(ks *servingv1alpha1.KnativeServing, c client.Client) error {
 	Configure(ks, "network", "ingress.class", "kourier.ingress.networking.knative.dev")
 	return nil
 }
 
-// configure egress
-func egress(ks *servingv1alpha1.KnativeServing, c client.Client) error {
-	networkConfig := &configv1.Network{}
-	if err := c.Get(context.TODO(), client.ObjectKey{Name: "cluster"}, networkConfig); err != nil {
-		if !meta.IsNoMatchError(err) {
-			return err
-		}
-		log.Info("No OpenShift cluster network config available")
-		return nil
-	}
-	network := strings.Join(networkConfig.Spec.ServiceNetwork, ",")
-	Configure(ks, "network", "istio.sidecar.includeOutboundIPRanges", network)
-	return nil
-}
-
 // configure ingress
 func ingress(ks *servingv1alpha1.KnativeServing, c client.Client) error {
 	ingressConfig := &configv1.Ingress{}
 	if err := c.Get(context.TODO(), client.ObjectKey{Name: "cluster"}, ingressConfig); err != nil {
 		if !meta.IsNoMatchError(err) {
-			return err
+			return fmt.Errorf("failed to fetch ingress config: %w", err)
 		}
 		log.Info("No OpenShift ingress config available")
 		return nil
