@@ -2,19 +2,24 @@ package knativeserving
 
 import (
 	"context"
+	"fmt"
 
 	"github.com/openshift-knative/serverless-operator/knative-operator/pkg/common"
 	"github.com/openshift-knative/serverless-operator/knative-operator/pkg/controller/knativeserving/consoleclidownload"
 	"github.com/openshift-knative/serverless-operator/knative-operator/pkg/controller/knativeserving/kourier"
-	"github.com/openshift-knative/serverless-operator/knative-operator/pkg/controller/knativeserving/servicemesh"
-	obsolete "github.com/openshift-knative/serverless-operator/serving/operator/pkg/apis/serving/v1alpha1"
 	"github.com/operator-framework/operator-sdk/pkg/k8sutil"
 	"github.com/operator-framework/operator-sdk/pkg/predicate"
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/equality"
 	"k8s.io/apimachinery/pkg/api/errors"
+	"k8s.io/apimachinery/pkg/api/meta"
+	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/runtime"
+	"k8s.io/apimachinery/pkg/runtime/schema"
+	"k8s.io/apimachinery/pkg/selection"
 	"k8s.io/apimachinery/pkg/types"
+	"k8s.io/apimachinery/pkg/util/sets"
+	"knative.dev/pkg/apis/istio/v1alpha3"
 	servingv1alpha1 "knative.dev/serving-operator/pkg/apis/serving/v1alpha1"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/controller"
@@ -25,6 +30,12 @@ import (
 	"sigs.k8s.io/controller-runtime/pkg/source"
 )
 
+const (
+	routeLabelKey     = "serving.knative.dev/route"
+	ingressClassKey   = "networking.knative.dev/ingress.class"
+	istioIngressClass = "istio.ingress.networking.knative.dev"
+)
+
 var log = common.Log.WithName("controller")
 
 // Add creates a new KnativeServing Controller and adds it to the Manager. The Manager will set fields on the Controller
@@ -52,6 +63,36 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 		return err
 	}
 
+	// Watch for Kourier resources.
+	manifest, err := kourier.RawManifest(mgr.GetClient())
+	if err != nil {
+		return err
+	}
+	resources := manifest.Resources()
+
+	gvkToKourier := make(map[schema.GroupVersionKind]runtime.Object)
+	for i := range resources {
+		resource := &resources[i]
+		gvkToKourier[resource.GroupVersionKind()] = resource
+	}
+
+	for _, t := range gvkToKourier {
+		err = c.Watch(&source.Kind{Type: t}, &handler.EnqueueRequestsFromMapFunc{
+			ToRequests: handler.ToRequestsFunc(func(obj handler.MapObject) []reconcile.Request {
+				annotations := obj.Meta.GetAnnotations()
+				ownerNamespace := annotations[kourier.OwnerNamespace]
+				ownerName := annotations[kourier.OwnerName]
+				if ownerNamespace != "" && ownerName != "" {
+					return []reconcile.Request{{
+						NamespacedName: types.NamespacedName{Namespace: ownerNamespace, Name: ownerName},
+					}}
+				}
+				return nil
+			})})
+		if err != nil {
+			return err
+		}
+	}
 	return nil
 }
 
@@ -72,8 +113,8 @@ func (r *ReconcileKnativeServing) Reconcile(request reconcile.Request) (reconcil
 	reqLogger.Info("Reconciling KnativeServing")
 
 	// Fetch the KnativeServing instance
-	instance := &servingv1alpha1.KnativeServing{}
-	err := r.client.Get(context.TODO(), request.NamespacedName, instance)
+	original := &servingv1alpha1.KnativeServing{}
+	err := r.client.Get(context.TODO(), request.NamespacedName, original)
 	if err != nil {
 		if errors.IsNotFound(err) {
 			return reconcile.Result{}, nil
@@ -81,39 +122,89 @@ func (r *ReconcileKnativeServing) Reconcile(request reconcile.Request) (reconcil
 		return reconcile.Result{}, err
 	}
 
-	if instance.GetDeletionTimestamp() != nil {
-		return reconcile.Result{}, r.delete(instance)
+	if original.GetDeletionTimestamp() != nil {
+		return reconcile.Result{}, r.delete(original)
 	}
 
+	instance := original.DeepCopy()
+	reconcileErr := r.reconcileKnativeServing(instance)
+
+	if !equality.Semantic.DeepEqual(original.Status, instance.Status) {
+		if err := r.client.Status().Update(context.TODO(), instance); err != nil {
+			return reconcile.Result{}, fmt.Errorf("failed to update status: %w", err)
+		}
+	}
+	return reconcile.Result{}, reconcileErr
+}
+
+func (r *ReconcileKnativeServing) reconcileKnativeServing(instance *servingv1alpha1.KnativeServing) error {
 	stages := []func(*servingv1alpha1.KnativeServing) error{
 		r.configure,
 		r.ensureFinalizers,
 		r.ensureCustomCertsConfigMap,
 		r.createConsoleCLIDownload,
 		r.installKourier,
-		r.uninstallServiceMesh,
+		r.updateDeployment,
+		r.deleteVirtualService,
 	}
 	for _, stage := range stages {
 		if err := stage(instance); err != nil {
-			return reconcile.Result{}, err
+			return err
 		}
 	}
-
-	return reconcile.Result{}, nil
+	return nil
 }
 
 // configure default settings for OpenShift
 func (r *ReconcileKnativeServing) configure(instance *servingv1alpha1.KnativeServing) error {
 	before := instance.DeepCopy()
-	log.Info("Configuring KnativeServing for OpenShift")
 	if err := common.Mutate(instance, r.client); err != nil {
 		return err
 	}
 	if equality.Semantic.DeepEqual(before, instance) {
 		return nil
 	}
+
 	// Only apply the update if something changed.
-	return r.client.Update(context.TODO(), instance)
+	log.Info("Updating KnativeServing with mutated state for Openshift")
+	if err := r.client.Update(context.TODO(), instance); err != nil {
+		return fmt.Errorf("failed to update KnativeServing with mutated state: %w", err)
+	}
+	return nil
+}
+
+// deleteVirtualService removes obsoleted VirtualServices.
+func (r *ReconcileKnativeServing) deleteVirtualService(instance *servingv1alpha1.KnativeServing) error {
+	labelSelector := labels.NewSelector()
+	req, err := labels.NewRequirement(routeLabelKey, selection.Exists, nil)
+	if err != nil {
+		return fmt.Errorf("failed to create requirement for label: %w", err)
+	}
+	listOpts := &client.ListOptions{LabelSelector: labelSelector.Add(*req)}
+	list := &v1alpha3.VirtualServiceList{}
+	ctx := context.TODO()
+	if err := r.client.List(ctx, listOpts, list); err != nil {
+		if meta.IsNoMatchError(err) {
+			// VirtualService CRD is not installed.
+			return nil
+		}
+		return err
+	}
+	for i := range list.Items {
+		vs := &list.Items[i]
+		if vs.GetAnnotations()[ingressClassKey] == istioIngressClass {
+			log.Info(fmt.Sprintf("deleting VirtualService %s/%s", vs.GetNamespace(), vs.GetName()))
+			if err := r.client.Delete(ctx, vs); err != nil {
+				return fmt.Errorf("failed to delete VirtualService %s/%s: %w", vs.GetNamespace(), vs.GetName(), err)
+			}
+		}
+	}
+	return nil
+}
+
+// updateDeployment updates Knative controller deployment
+func (r *ReconcileKnativeServing) updateDeployment(instance *servingv1alpha1.KnativeServing) error {
+	return common.ApplyProxySettings(instance, r.client)
 }
 
 // set a finalizer to clean up service mesh when instance is deleted
@@ -123,6 +214,7 @@ func (r *ReconcileKnativeServing) ensureFinalizers(instance *servingv1alpha1.Kna
 			return nil
 		}
 	}
+	log.Info("Adding finalizer")
 	instance.SetFinalizers(append(instance.GetFinalizers(), finalizerName()))
 	return r.client.Update(context.TODO(), instance)
 }
@@ -143,8 +235,10 @@ func (r *ReconcileKnativeServing) ensureCustomCertsConfigMap(instance *servingv1
 			if err := controllerutil.SetControllerReference(instance, cm, r.scheme); err != nil {
 				return err
 			}
+
+			log.Info("Creating Custom Certs Config Map")
 			if err = r.client.Create(ctx, cm); err != nil {
-				return err
+				return fmt.Errorf("failed to create custom certs config map: %w", err)
 			}
 			return nil
 		}
@@ -159,11 +253,6 @@ func (r *ReconcileKnativeServing) installKourier(instance *servingv1alpha1.Knati
 	return kourier.Apply(instance, r.client, r.scheme)
 }
 
-// Uninstall obsolete SMCP deployed by previous version
-func (r *ReconcileKnativeServing) uninstallServiceMesh(instance *servingv1alpha1.KnativeServing) error {
-	return servicemesh.Delete(instance, r.client)
-}
-
 // createConsoleCLIDownload creates CR for kn CLI download link
 func (r *ReconcileKnativeServing) createConsoleCLIDownload(instance *servingv1alpha1.KnativeServing) error {
 	return consoleclidownload.Create(instance, r.client)
@@ -171,33 +260,40 @@ func (r *ReconcileKnativeServing) createConsoleCLIDownload(instance *servingv1al
 
 // general clean-up, mostly resources in different namespaces from servingv1alpha1.KnativeServing.
 func (r *ReconcileKnativeServing) delete(instance *servingv1alpha1.KnativeServing) error {
-	if len(instance.GetFinalizers()) == 0 || instance.GetFinalizers()[0] != finalizerName() {
+	finalizer := finalizerName()
+	finalizers := sets.NewString(instance.GetFinalizers()...)
+
+	if !finalizers.Has(finalizer) {
+		log.Info("Finalizer has already been removed, nothing to do")
 		return nil
 	}
 
-	// Try to delete the "old" CR, if it's still there. Ignore any errors.
-	old := &obsolete.KnativeServing{}
-	if err := r.client.Get(context.TODO(), types.NamespacedName{Namespace: instance.Namespace, Name: instance.Name}, old); err == nil {
-		// Oh, it actually still exists. Remove it, but ignore all errors!
-		r.client.Delete(context.TODO(), old)
-	}
-
-	if err := kourier.Delete(instance, r.client); err != nil {
-		return err
+	log.Info("Running cleanup logic")
+	log.Info("Deleting kourier")
+	if err := kourier.Delete(instance, r.client, r.scheme); err != nil {
+		return fmt.Errorf("failed to delete kourier: %w", err)
 	}
 
+	log.Info("Deleting ConsoleCLIDownload")
 	if err := consoleclidownload.Delete(instance, r.client); err != nil {
-		log.Info("Failed to delete ConsoleCLIDownload CR for kn..")
-		return err
+		return fmt.Errorf("failed to delete ConsoleCLIDownload: %w", err)
 	}
 
-	// The deletionTimestamp might've changed. Fetch the resource again.
+	// The above might take a while, so we refetch the resource again in case it has changed.
 	refetched := &servingv1alpha1.KnativeServing{}
 	if err := r.client.Get(context.TODO(), types.NamespacedName{Namespace: instance.Namespace, Name: instance.Name}, refetched); err != nil {
-		return err
+		return fmt.Errorf("failed to refetch KnativeServing: %w", err)
 	}
-	refetched.SetFinalizers(refetched.GetFinalizers()[1:])
-	return r.client.Update(context.TODO(), refetched)
+
+	// Update the refetched finalizer list.
+	finalizers = sets.NewString(refetched.GetFinalizers()...)
+	finalizers.Delete(finalizer)
+	refetched.SetFinalizers(finalizers.List())
+
+	if err := r.client.Update(context.TODO(), refetched); err != nil {
+		return fmt.Errorf("failed to update KnativeServing with removed finalizer: %w", err)
+	}
+	return nil
 }
 
 func finalizerName() string {
